# pylint: disable=invalid-name

import importlib
from dataclasses import dataclass
from types import TracebackType
from typing import Any, Optional, Type

from .file_tree import FileTree
from .unsupported import block_unsupported


@dataclass
class MockFileTree:
    os: Any
    tree: FileTree
    safe: bool = False

    def __enter__(self) -> "MockFileTree":
        self.apply()
        return self

    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> None:
        self.restore()

    def apply(self) -> None:
        if self.safe:
            block_unsupported(self.os)

        self.os.listdir = self.tree.listdir
        self.os.path.exists = self.tree.path_exists
        self.os.path.isdir = self.tree.path_isdir
        self.os.path.isfile = self.tree.path_isfile

    def restore(self) -> None:
        importlib.reload(self.os)
        importlib.reload(self.os.path)



import os
from typing import List


def path_base(path: str) -> str:
    """
    >>> path_base('/dir1/dir2/file1.ext')
    ''

    >>> path_base('dir1/dir2/file1.ext')
    'dir1'
    """
    return os.path.normpath(path).split(os.sep)[0]


def path_from_base(path: str) -> str:
    """
    >>> path_from_base('/dir1/dir2/file1.ext')
    'dir1/dir2/file1.ext'

    >>> path_from_base('dir1/dir2/file1.ext')
    'dir2/file1.ext'
    """
    return path[1:] if path_base(path) == "" else os.path.relpath(path, path_base(path))


def child_paths(base: str, paths: List[str]) -> List[str]:
    """
    >>> child_paths('dir1', ['file1', 'dir1/file2'])
    ['file2']
    """
    return [
        path_from_base(j)
        for j in paths
        if path_base(j) == base and path_from_base(j) != "."
    ]



from typing import Any


class SubtreeNotFound(Exception):

    # We type the FileTree as Any here as a tradeoff. It lets us spearate out exceptions
    # into a standalone module and avoid a circular dependency without any import hacks.
    def __init__(self, path: str, tree: Any) -> None:
        super().__init__(
            f"Could not find the path: '{path}' in the file tree: '{tree}'."
        )



# pylint: disable=invalid-name
# pylint: disable=unused-variable
from typing import Any, NoReturn


def unsupported(*args: Any, **kwargs: Any) -> NoReturn:
    raise NotImplementedError()


# We can't do this as a loop as we're redefining the variables.
def block_unsupported(os: Any) -> None:
    os.access = unsupported
    os.chdir = unsupported
    os.chflags = unsupported
    os.chmod = unsupported
    os.chown = unsupported
    os.chroot = unsupported
    os.lchflags = unsupported
    os.lchmod = unsupported
    os.lchown = unsupported
    os.lstat = unsupported
    os.mkdir = unsupported
    os.mkfifo = unsupported
    os.mknod = unsupported
    os.open = unsupported
    os.path.getatime = unsupported
    os.path.getctime = unsupported
    os.path.getmtime = unsupported
    os.path.getsize = unsupported
    os.path.islink = unsupported
    os.path.ismount = unsupported
    os.path.lexists = unsupported
    os.pathconf = unsupported
    os.readlink = unsupported
    os.remove = unsupported
    os.rmdir = unsupported
    os.scandir = unsupported
    os.stat = unsupported
    os.statvfs = unsupported
    os.truncate = unsupported
    os.unlink = unsupported
    os.utime = unsupported



from functools import reduce
from typing import Any, Dict, List, cast

from setuptools import setup

configuration = {
    "packages": ['mock_file_tree'],
    "package_dir": {'mock_file_tree': 'src'},
    "install_requires": [
        "dataclasses",
        "types-dataclasses",
    ],
    "extras_require": {
        "linters": [
            "mypy",
            "pylint",
            "bandit",
        ],
        "formatters": [
            "autoflake",
            "black",
            "isort",
        ],
        "tests": ["tox", "coverage", "pytest"],
        "docs": ["quickdocs"],
        "publishers": [
            "twine",
            "wheel",
            "bump2version",
        ],
    },
}

if __name__ == "__main__":
    extras_require = cast(Dict[Any, List[Any]], configuration["extras_require"])

    merge_lists = lambda acc, x: list(set(acc) | (set(x)))

    configuration["extras_require"] = dict(
        **extras_require, **{"all": reduce(merge_lists, extras_require.values())}
    )

    setup(**configuration)



import os
from dataclasses import dataclass
from typing import List

from ..utils.paths import child_paths, path_base, path_from_base
from .exceptions import SubtreeNotFound


@dataclass
class FileTree:
    base: str
    children: List["FileTree"]

    @classmethod
    def from_paths(cls, *paths: str, base: str = ".") -> "FileTree":
        return cls(
            base,
            [
                cls.from_paths(*child_paths(i, list(paths)), base=i)
                for i in set(path_base(j) for j in paths)
            ],
        )

    def is_descendant(self, path: str) -> bool:
        path = os.path.normpath(path)
        return (
            path == self.base
            or path in [i.base for i in self.children]
            or any(i.is_descendant(path_from_base(path)) for i in self.children)
        )

    def get_descendant(self, path: str) -> "FileTree":
        if not self.is_descendant(path):
            raise SubtreeNotFound(path, self)

        path = os.path.normpath(path)

        if path == self.base:
            return self

        if path in [i.base for i in self.children]:
            return next(filter(lambda x: x.base == path, self.children))

        return next(
            i.get_descendant(path_from_base(path))
            for i in self.children
            if i.is_descendant(path_from_base(path))
        )

    def listdir(self, path: str = ".") -> List[str]:
        return [i.base for i in self.get_descendant(path).children]

    def path_exists(self, path: str) -> bool:
        return self.is_descendant(path)

    def path_isdir(self, path: str) -> bool:
        return self.is_descendant(path) and len(self.get_descendant(path).children) > 0

    def path_isfile(self, path: str) -> bool:
        return self.is_descendant(path) and len(self.get_descendant(path).children) == 0
